@@ -4,19 +4,22 @@ import base64
 import logging
 import os
 import random
+import re
+import warnings
 from datetime import datetime
 from datetime import timedelta
 from datetime import timezone
+from decimal import Decimal
 from io import BytesIO
 from tempfile import NamedTemporaryFile
 from typing import BinaryIO
 from uuid import uuid4
 
-import pytz
 from django.conf import settings
 from django.core import management
 from django.core.files import File
 from django.core.files.storage import Storage
+from django.db import connection
 from django.db import models
 from django.db.models import Count
 from django.db.models import Sum
@@ -30,6 +33,8 @@ from OpenSSL.crypto import X509
 from OpenSSL.crypto import load_certificate
 from zeep.exceptions import Fault
 
+from django_afip.clients import TZ_AR
+
 from . import clients
 from . import crypto
 from . import exceptions
@@ -37,7 +42,6 @@ from . import parsers
 from . import serializers
 
 logger = logging.getLogger(__name__)
-TZ_AR = pytz.timezone(pytz.country_timezones["ar"][0])
 
 # http://www.afip.gov.ar/afip/resol1415_anexo2.html
 VAT_CONDITIONS = (
@@ -236,6 +240,35 @@ class VatType(GenericAfipType):
     See the AFIP's documentation for details on each VAT type.
     """
 
+    @property
+    def as_decimal(self) -> Decimal:
+        """Return this VatType as a Decimal.
+
+        Parses the percent amount from the ``description`` field. This number is usable
+        when calculating the Vat for entries which have this type. If Vat is 21%, then
+        the returned value is ``Decimal("0.21")``.
+
+        Assuming that an item pays 21% vat, when using a net price for the calculation,
+        the following are all correct::
+
+            total_price = net_price * Decimal(1.21)
+            vat = net_price * Decimal(0.21)
+            vat = total_price - net_price
+
+        If using the total price, this approach should be used (this can be derived from
+        the above)::
+
+            net_price = total_price / Decimal(1.21)
+            vat = total_price - net_price
+
+        Keep in mind that AFIP requires the usage of "round half even", which is what
+        Python's ``Decimal`` class uses by default (See ``decimal.ROUND_HALF_EVEN``).
+        """
+        match = re.match(r"^([0-9]{1,2}\.?[0-9]{0,2})%$", self.description)
+        if not match:
+            raise ValueError("The description for this VatType is not a percentage.")
+        return Decimal(match.groups()[0]) / 100
+
     objects = GenericAfipTypeManager("FEParamGetTiposIva", "IvaTipo")
 
     class Meta:
@@ -284,6 +317,15 @@ class TaxPayer(models.Model):
     - logo
     """
 
+    # XXX: Split this into TaxPayer and Credentials
+    # The former has a unique CUIT, which will be its natural key.
+    # The latter has the key and cert and alike.
+    #
+    # For the migration path:
+    # - Make sure that if there's 1:1 TaxPayer:Credentials,
+    #   everything continues to work as it does now.
+    # - Provide data migrations
+
     name = models.CharField(
         _("name"),
         max_length=32,
@@ -316,7 +358,7 @@ class TaxPayer(models.Model):
     certificate_expiration = models.DateTimeField(
         _("certificate expiration"),
         editable=False,
-        null=True,  # Either no cert, or and old TaxPayer
+        null=True,  # Either no cert, or an old TaxPayer.
         help_text=_(
             "Stores expiration for the current certificate.<br>Note that this "
             "field is updated pre-save, so the value may be invalid for unsaved models."
@@ -337,7 +379,10 @@ class TaxPayer(models.Model):
 
     @property
     def logo_as_data_uri(self) -> str:
-        """This TaxPayer's logo as a data uri."""
+        """This TaxPayer's logo as a data uri.
+
+        This can be used to embed the image into an HTML or PDF file.
+        """
         _, ext = os.path.splitext(self.logo.file.name)
         with self.logo.open() as f:
             data = base64.b64encode(f.read())
@@ -360,9 +405,7 @@ class TaxPayer(models.Model):
         return load_certificate(FILETYPE_PEM, self.certificate.read())
 
     def get_certificate_expiration(self) -> datetime | None:
-        """
-        Gets the certificate expiration from the certificate
-
+        """Return the certificate expiration from the current certificate
 
         Gets the certificate expiration from the certificate file. Note that
         this value is stored into ``certificate_expiration`` when an instance
@@ -380,8 +423,7 @@ class TaxPayer(models.Model):
         return dt.replace(tzinfo=timezone.utc)
 
     def generate_key(self, force=False) -> bool:
-        """
-        Creates a key file for this TaxPayer
+        """Creates a key file for this TaxPayer
 
         Creates a key file for this TaxPayer if it does not have one, and
         immediately saves it.
@@ -403,11 +445,15 @@ class TaxPayer(models.Model):
         return True
 
     def generate_csr(self, basename="djangoafip") -> BinaryIO:
-        """
-        Creates a CSR for this TaxPayer's key
+        """Creates a CSR with this TaxPayer's key
 
-        Creates a file-like object that contains the CSR which can be used to
-        request a new certificate from AFIP.
+        The CSR (certificate signing request) can be used to request a new certificate
+        via AFIP's website. After generating a new CSR, it should be manually uploaded
+        to AFIP's website, and a new certificate will be returned. That certificate
+        should be uploaded to the ``certificate`` field.
+
+        It is safe to use with when renovating expired certificates on production
+        systems.
         """
         csr = BytesIO()
         crypto.create_csr(
@@ -421,13 +467,21 @@ class TaxPayer(models.Model):
         return csr
 
     def create_ticket(self, service: str) -> AuthTicket:
-        """Create an AuthTicket for a given service."""
+        """Create an AuthTicket for a given service.
+
+        Tickets are saved to the database. It is recommended to use the
+        :meth:`~.TaxPayer.get_or_create_ticket` method instead.
+        """
         ticket = AuthTicket(owner=self, service=service)
         ticket.authorize()
         return ticket
 
     def get_ticket(self, service: str) -> AuthTicket | None:
-        """Return an existing AuthTicket for a given service, if any."""
+        """Return an existing AuthTicket for a given service, if any.
+
+        It is recommended to use the :meth:`~.TaxPayer.get_or_create_ticket` method
+        instead.
+        """
         return self.auth_tickets.filter(
             expires__gt=datetime.now(timezone.utc),
             service=service,
@@ -470,7 +524,7 @@ class TaxPayer(models.Model):
                     owner=self,
                     defaults={
                         "issuance_type": pos_data.EmisionTipo,
-                        "blocked": pos_data.Bloqueado == "N",
+                        "blocked": pos_data.Bloqueado == "S",
                         "drop_date": parsers.parse_date(pos_data.FchBaja),
                     },
                 )
@@ -522,9 +576,11 @@ class PointOfSales(models.Model):
     number = models.PositiveSmallIntegerField(
         _("number"),
     )
+    # AFIP has replied that this field may be up to 200bytes,
+    # so 200 characters should always be more than enough.
     issuance_type = models.CharField(
         _("issuance type"),
-        max_length=24,
+        max_length=200,
         help_text="Indicates if this POS emits using CAE and CAEA.",
     )
     blocked = models.BooleanField(
@@ -664,7 +720,7 @@ class AuthTicket(models.Model):
     )
     service = models.CharField(
         _("service"),
-        max_length=6,
+        max_length=34,
         help_text=_("Service for which this ticket has been authorized."),
     )
 
@@ -681,6 +737,9 @@ class AuthTicket(models.Model):
     SIGN_XPATH = "/loginTicketResponse/credentials/sign"
 
     def __create_request_xml(self) -> bytes:
+        """Create a new ticket request XML
+
+        This is the payload we sent to AFIP to request a new ticket."""
         request_xml = E.loginTicketRequest(
             {"version": "1.0"},
             E.header(
@@ -735,6 +794,10 @@ class AuthTicket(models.Model):
 class ReceiptQuerySet(models.QuerySet):
     """The default queryset obtains when querying via :class:`~.ReceiptManager`."""
 
+    # This private flag is provided only to disable the durability checks in tests.
+    # Inspired by Django's flag of the same name for `Atomic`.
+    _ensure_durability = True
+
     def _assign_numbers(self) -> None:
         """Assign numbers in preparation for validating these receipts.
 
@@ -783,24 +846,47 @@ class ReceiptQuerySet(models.QuerySet):
         """Validate all receipts matching this queryset.
 
         Note that, due to how AFIP implements its numbering, this method is not
-        thread-safe, or even multiprocess-safe.
+        thread-safe, or even multiprocess-safe. You MAY however, call this method
+        concurrently for receipts from different :class:`~.PointOfSales`.
 
-        Because of this, it is possible that not all instances matching this
-        queryset are validated properly. Obviously, only successfully validated
-        receipts will be updated.
+        It is possible that not all instances matching this queryset are validated
+        properly. This method is written in a way that the database will always remain
+        in a consistent state.
 
-        Returns a list of errors as returned from AFIP's webservices. An
-        exception is not raised because partial failures are possible.
+        Only successfully validated receipts will marked as such. This method takes care
+        of saving all changes to database.
 
-        Receipts that succesfully validate will have a
-        :class:`~.ReceiptValidation` object attatched to them with a validation
-        date and CAE information.
+        Returns a list of errors as returned from AFIP's webservices. When AFIP returns
+        a failure response, an exception is not raised because partial failures are
+        possible. Network issues (e.g.: DNS failure) _will_ raise an exception.
+
+        Receipts that successfully validate will have a :class:`~.ReceiptValidation`
+        object attached to them with a validation date and CAE information.
 
         Already-validated receipts are ignored.
 
         Attempting to validate an empty queryset will simply return an empty
         list.
+
+        This method takes the following steps:
+
+            - Assigns numbers to all receipts.
+            - Saves the assigned numbers to the database.
+            - Sends the receipts to AFIP.
+            - Saves the results into the local DB.
+
+        Should execution be interrupted (e.g.: a power failure), receipts will have been
+        saved with their number. In this case, the ``revalidate`` method should be used,
+        to determine if they have been registered by AFIP, or if the interruption
+        happened before sending them.
+
+        Calling this method inside a transaction will raise ``RuntimeError``, since
+        doing so risks leaving the database in an inconsistent state should there be any
+        fatal interruptions. In particular, the receipt numbers will not have been
+        saved, so it would be impossible to recover from the incomplete operation.
         """
+        if self._ensure_durability and connection.in_atomic_block:
+            raise RuntimeError("This function cannot be called within a transaction")
         # Skip any already-validated ones:
         qs = self.filter(validation__isnull=True).check_groupable()
         if qs.count() == 0:
@@ -857,7 +943,7 @@ class ReceiptQuerySet(models.QuerySet):
 class ReceiptManager(models.Manager):
     """Default manager for the :class:`~.Receipt` class.
 
-    You should generally access this using ``Receipt.objects``.
+    This should be accessed using ``Receipt.objects``.
     """
 
     def fetch_last_receipt_number(
@@ -920,6 +1006,9 @@ class ReceiptManager(models.Manager):
             return None
 
     def get_queryset(self) -> ReceiptQuerySet:
+        """Return a new QuerySet object.
+
+        This always joins with :class:`~.ReceiptType`."""
         return ReceiptQuerySet(self.model, using=self._db).select_related(
             "receipt_type",
         )
@@ -938,6 +1027,12 @@ class Receipt(models.Model):
 
     If the taxpayer has taxes or pays VAT, you need to attach :class:`~.Tax`
     and/or :class:`~.Vat` instances to the Receipt.
+
+    Application code SHOULD NOT set the `receipt_number` code. It will be set by
+    :meth:`~.Receipt.validate()` internally. When writing code outside `django-afip`,
+    this should be considered read-only. The sole exception is importing
+    previously-validated receipts from another database.
+    `
     """
 
     point_of_sales = models.ForeignKey(
@@ -1116,14 +1211,25 @@ class Receipt(models.Model):
     def validate(self, ticket: AuthTicket = None, raise_=False) -> list[str]:
         """Validates this receipt.
 
-        This is a shortcut to :class:`~.ReceiptQuerySet`'s method of the same
-        name. Calling this validates only this instance.
+        This is a shortcut to :meth:`~.ReceiptQuerySet.validate`. See the documentation
+        for that method for details. Calling this validates only this instance.
 
-        :param AuthTicket ticket: Use this ticket. If None, one will be loaded
-            or created automatically.
-        :param bool raise_: If True, an exception will be raised when
-            validation fails.
+
+        .. versionchanged:: 11
+
+            The ``raise_`` flag has been deprecated.
+
+        :param ticket: Use this ticket. If None, one will be loaded or created
+            automatically.
+        :param raise_: If True, an exception will be raised when validation fails.
         """
+        if raise_:
+            warnings.warn(
+                "The raise_ flag is deprecated and will be removed in django_afip 12.",
+                DeprecationWarning,
+                stacklevel=2,
+            )
+
         # XXX: Maybe actually have this sortcut raise an exception?
         rv = Receipt.objects.filter(pk=self.pk).validate(ticket)
         # Since we're operating via a queryset, this instance isn't properly
@@ -1208,8 +1314,8 @@ class ReceiptPDFManager(models.Manager):
         All attributes will be completed with the information for the relevant
         :class:`~.PointOfSales` instance.
 
-        :param Receipt receipt: The receipt for the PDF which will be
-            generated.
+        :param receipt: The receipt for the PDF which will be generated.
+        :param **kwargs: Passed directly to the :class:`~.ReceiptPDF` constructor.
         """
         pdf = ReceiptPDF.objects.create(
             receipt=receipt,
