@@ -6,12 +6,12 @@ from django_afip import models
 
 
 @receiver(pre_save, sender=models.TaxPayer)
-def update_certificate_expiration(sender, instance, **kwargs):
+def update_certificate_expiration(sender, instance: models.TaxPayer, **kwargs):
     if instance.certificate:
         instance.certificate_expiration = instance.get_certificate_expiration()
 
 
 @receiver(post_save, sender=models.ReceiptPDF)
-def generate_receipt_pdf(sender, instance, **kwargs):
+def generate_receipt_pdf(sender, instance: models.ReceiptPDF, **kwargs):
     if not instance.pdf_file and instance.receipt.is_validated:
         instance.save_pdf(save_model=True)
